==> bonobo-0.28.tar.gz: bonobo-0.28/monikers/bonobo-moniker-gunzip.c <==
/*
 * bonobo-moniker-gunzip.c: gunzip based Moniker
 *
 * Copyright (c) 2000 Helix Code, Inc.
 *
 * Author:
 *   Joe Shaw (joe@helixcode.com)
 *
 */

#include <config.h>
#include <gnome.h>
#include <liboaf/liboaf.h>
#include <bonobo/bonobo.h>
#include <bonobo/bonobo-stream-memory.h>
#include <zlib.h>

#include "bonobo-moniker-gunzip.h"

#define PREFIX_LEN (sizeof ("gunzip:") - 1)

static int running_objects = 0;
static BonoboGenericFactory *gunzip_factory = NULL;

static BonoboMonikerClass *bonobo_moniker_gunzip_parent_class;

static Bonobo_Moniker 
gunzip_parse_display_name(BonoboMoniker *moniker,
			  Bonobo_Moniker parent,
			  const CORBA_char *name,
			  CORBA_Environment *ev)
{
	int i;

	g_return_val_if_fail(moniker != NULL, CORBA_OBJECT_NIL);
	g_return_val_if_fail(strlen(name) >= PREFIX_LEN, CORBA_OBJECT_NIL);

	bonobo_moniker_set_parent(moniker, parent, ev);

	i = bonobo_moniker_util_seek_std_separator(name, 0);

	bonobo_moniker_set_name(moniker, name, i);

	if (name[i])
		return bonobo_moniker_util_new_from_name_full(
			bonobo_object_corba_objref(BONOBO_OBJECT(moniker)),
			&name[i], ev);
	else
		return bonobo_object_dup_ref(
			bonobo_object_corba_objref(BONOBO_OBJECT(moniker)), 
			ev);
} /* gunzip_parse_display_name */

/*
 * Count number of bytes to skip at start of buf
 */
static int gz_magic[2] = {0x1f, 0x8b};
/* gzip flag byte */
#define GZ_ASCII_FLAG   0x01 /* bit 0 set: file probably ascii text */
#define GZ_HEAD_CRC     0x02 /* bit 1 set: header CRC present */
#define GZ_EXTRA_FIELD  0x04 /* bit 2 set: extra field present */
#define GZ_ORIG_NAME    0x08 /* bit 3 set: original file name present */
#define GZ_COMMENT      0x10 /* bit 4 set: file comment present */
#define GZ_RESERVED     0xE0 /* bits 5..7: reserved */

static int
count_gzip_header (guint8 *buf, guint32 input_length)
{
	int method, flags;
	guint8 *s = buf;
	guint32 left_len = input_length;
	
	if (left_len < 4) return -1;
	if (*s++ != gz_magic[0] || *s++ != gz_magic[1]) {
		return -2;
	}
	
	method = *s++;
	flags = *s++;
	left_len -= 4;
	
	if (method != Z_DEFLATED || (flags & GZ_RESERVED) != 0) {
		/* If it's not deflated, or the reserved isn't 0 */
		return -3;
	}
	
	/* Skip time, xflags, OS code */
	if (left_len < 6) return -4;
	s += 6;
	left_len -= 6;
	
	if (flags & GZ_EXTRA_FIELD) {
		unsigned int len;
		if (left_len < 2) return -5;
		len = (unsigned int)(*s++);
		len += ((unsigned int)(*s++)) << 8;
		if (left_len < len) return -6;
		s += len;
		left_len -= len;
	}
	
	/* Skip filename */
	if (flags & GZ_ORIG_NAME) {
		while (--left_len != 0 && *s++ != '\0') ;
		if (left_len == 0) return -7;
	}
	/* Skip comment */
	if (flags & GZ_COMMENT) {
		while (--left_len != 0 && *s++ != '\0') ;
		if (left_len == 0) return -7;
	}
	/* Skip CRC */
	if (flags & GZ_HEAD_CRC) {
		if (left_len < 2) return -7;
		s += 2;
		left_len -= 2;
	}
	
	return input_length - left_len;
} /* count_gzip_header */

static char *
uncompress_memory(char *input_buffer, int size, int *out_size)
{
	z_stream zs;
	gchar *outbuf = NULL;
	GByteArray *data;
	int data_len = 0;
	int zret;
	int gzip_hdr;

	g_return_val_if_fail(input_buffer, NULL);

	data = g_byte_array_new ();

	gzip_hdr = count_gzip_header (input_buffer, size);
	if (gzip_hdr < 0) {
		/* This is not a gzipped stream. Just return the stuff */
		g_byte_array_free(data, FALSE);
		outbuf = g_memdup(input_buffer, size);
		if (out_size)
			*out_size = size;
		return outbuf;
	}

	zs.next_in = input_buffer + gzip_hdr;
	zs.avail_in = size - gzip_hdr;
	zs.zalloc = NULL;
	zs.zfree = NULL;
	zs.opaque = NULL;
	
	outbuf = g_malloc (10000);
	zs.next_out = outbuf;
	zs.avail_out = 10000;

	/* Negative inflateinit is magic to tell zlib that there is no
	 * zlib header */
	inflateInit2 (&zs, -MAX_WBITS);

	while (1) {
		zret = inflate (&zs, Z_SYNC_FLUSH);
		if (zret != Z_OK && zret != Z_STREAM_END)
			break;

		g_byte_array_append(data, outbuf, 10000 - zs.avail_out);
		data_len += 10000 - zs.avail_out;
		zs.next_out = outbuf;
		zs.avail_out = 10000;

		if (zret == Z_STREAM_END)
			break;
	}
	
	if (zret != Z_STREAM_END) {
		g_warning ("libz inflate failed! (%d)", zret);
	}
	
	inflateEnd (&zs);

	g_free (outbuf);

	outbuf = data->data;
	if (out_size)
		*out_size = data->len;
	g_byte_array_free(data, FALSE);

	return outbuf;
} /* uncompress_memory */


static Bonobo_Unknown
gunzip_resolve(BonoboMoniker *moniker,
	       const Bonobo_ResolveOptions *options,
	       const CORBA_char *requested_interface,
	       CORBA_Environment *ev)
{
	Bonobo_Moniker parent;
	const char *parent_name;
	Bonobo_Stream stream_in;
	int bytes;
	int len;
	Bonobo_Stream_iobuf *buffer;
	gboolean cont;
	char *ob;
	GByteArray *data;

	g_warning("Going to resolve the gunzip now");

	parent = bonobo_moniker_get_parent(moniker, ev);
	parent_name = bonobo_moniker_client_get_name(parent, ev);

	if (ev->_major != CORBA_NO_EXCEPTION)
		return CORBA_OBJECT_NIL;

	if (parent == CORBA_OBJECT_NIL) {
		g_warning("gunzip moniker with no parent!");
		CORBA_exception_set(
			ev, CORBA_USER_EXCEPTION,
			ex_Bonobo_Moniker_InterfaceNotFound, NULL);
		return CORBA_OBJECT_NIL;
	}

	stream_in = Bonobo_Moniker_resolve(
		parent, options, "IDL:Bonobo/Stream:1.0", ev);

	if (ev->_major != CORBA_NO_EXCEPTION)
		goto return_unref_parent;

	if (stream_in == CORBA_OBJECT_NIL) {
		g_warning("Failed to obtain a stream from the parent");
		CORBA_exception_set(
			ev, CORBA_USER_EXCEPTION,
			ex_Bonobo_Moniker_InterfaceNotFound, NULL);
		goto return_unref_parent;
	}

	/* This is necessary because while there is a 
	   bonobo_stream_client_read_string, there isn't any
	   bonobo_stream_client_read, and this compressed data isn't a
	   string. */
	data = g_byte_array_new();
	do {
		Bonobo_Stream_read(stream_in, 1, &buffer, ev);
		
		if (ev->_major != CORBA_NO_EXCEPTION) {
			g_byte_array_free(data, TRUE);
			goto return_unref_parent;
		}
		cont = buffer->_length ? TRUE : FALSE;
		if (buffer->_length) {
			data = g_byte_array_append(
				data, buffer->_buffer, buffer->_length);
		}
		CORBA_free(buffer);
	} while (cont);

	/* This function will return a memdup of the data we pass to it if it
	   can't decode the header. */
	ob = uncompress_memory(data->data, data->len, &len);
	g_byte_array_free(data, TRUE);
	/* We're going to write this data out. We don't use _write_string
	   because we can't guarantee that ob will be a string. */
	bonobo_stream_client_write(stream_in, ob, len, ev);
	g_free(ob);
#if 0
	stream = bonobo_stream_mem_create(ob, len, TRUE, FALSE);
	g_free(ob);

	if (!stream)
		goto return_unref_parent;

	return CORBA_Object_duplicate(
		bonobo_object_corba_objref(BONOBO_OBJECT(stream)), ev);
#endif

	return stream_in;

return_unref_parent:
	bonobo_object_release_unref(parent, ev);
	return CORBA_OBJECT_NIL;
} /* gunzip_resolve */

static void
bonobo_moniker_gunzip_class_init(BonoboMonikerGunzipClass *klass)
{
	BonoboMonikerClass *mclass = (BonoboMonikerClass *) klass;

	bonobo_moniker_gunzip_parent_class = gtk_type_class(
		bonobo_moniker_get_type());

	mclass->parse_display_name = gunzip_parse_display_name;
	mclass->resolve = gunzip_resolve;
} /* bonobo_moniker_gunzip_class_init */

GtkType
bonobo_moniker_gunzip_get_type(void)
{
	static GtkType type = 0;

	if (!type) {
		GtkTypeInfo info = {
			"BonoboMonikerGunzip",
			sizeof(BonoboMonikerGunzip),
			sizeof(BonoboMonikerGunzipClass),
			(GtkClassInitFunc) bonobo_moniker_gunzip_class_init,
			(GtkObjectInitFunc) NULL,
			NULL, /* reserved 1 */
			NULL, /* reserved 2 */
			(GtkClassInitFunc) NULL
		};

		type = gtk_type_unique(bonobo_moniker_get_type(), &info);
	}

	return type;
} /* bonobo_moniker_gunzip_get_type */

static BonoboMonikerGunzip *
bonobo_moniker_gunzip_new(void)
{
	BonoboMoniker *moniker;
	Bonobo_Moniker corba_moniker;

	moniker = gtk_type_new(bonobo_moniker_gunzip_get_type());
	corba_moniker = bonobo_moniker_corba_object_create(
		BONOBO_OBJECT(moniker));

	if (corba_moniker == CORBA_OBJECT_NIL) {
		bonobo_object_unref(BONOBO_OBJECT(moniker));
		return NULL;
	}

	if (!bonobo_object_construct(BONOBO_OBJECT(moniker), corba_moniker)) {
		bonobo_object_unref(BONOBO_OBJECT(moniker));
		return NULL;
	}

	return BONOBO_MONIKER_GUNZIP(moniker);
} /* bonobo_moniker_gunzip_new */

static void
gunzip_destroy_cb(BonoboMoniker *moniker)
{
	running_objects--;

	if (running_objects > 0)
		return;

	bonobo_object_unref(BONOBO_OBJECT(gunzip_factory));
	gtk_main_quit();
} /* gunzip_destroy_cb */

static BonoboObject *
bonobo_moniker_gunzip_factory(BonoboGenericFactory *this, void *closure)
{
	BonoboMonikerGunzip *gunzip_moniker = bonobo_moniker_gunzip_new();

	gtk_signal_connect(
		GTK_OBJECT(gunzip_moniker), "destroy",
		GTK_SIGNAL_FUNC(gunzip_destroy_cb), NULL);

	running_objects++;

	return BONOBO_OBJECT(gunzip_moniker);
} /* bonobo_moniker_gunzip_factory */

int
main(int argc, char *argv[])
{
	CORBA_Environment ev;
	CORBA_ORB orb = CORBA_OBJECT_NIL;

	CORBA_exception_init(&ev);

	gnome_init_with_popt_table(
		"gunzip-moniker", "0.0", argc, argv, oaf_popt_options, 0, NULL);
	orb = oaf_init(argc, argv);

	if (!bonobo_init(orb, CORBA_OBJECT_NIL, CORBA_OBJECT_NIL))
		g_error(_("Could not initialize Bonobo"));

	gunzip_factory = bonobo_generic_factory_new(
		"OAFIID:GunzipMonikerFactory:9ec1a8c4-ccae-4ad9-8e74-6622049b6d24",
		bonobo_moniker_gunzip_factory, NULL);

	bonobo_main();

	CORBA_exception_free(&ev);

	return 0;
} /* main */

==> bonobo-0.29.tar.gz: bonobo-0.29/monikers/bonobo-moniker-gunzip.c <==
/*
 * bonobo-moniker-gunzip.c: gunzip based Moniker
 *
 * Author:
 *   Joe Shaw (joe@helixcode.com)
 *
 * Copyright (c) 2000 Helix Code, Inc.
 */

#include <config.h>
#include <gnome.h>
#include <liboaf/liboaf.h>
#include <bonobo/bonobo.h>
#include <bonobo/bonobo-stream-memory.h>
#include <zlib.h>

#include "bonobo-moniker-gunzip.h"

#define PREFIX_LEN (sizeof ("gunzip:") - 1)

static int running_objects = 0;
static BonoboGenericFactory *gunzip_factory = NULL;

static BonoboMonikerClass *bonobo_moniker_gunzip_parent_class;

static Bonobo_Moniker 
gunzip_parse_display_name(BonoboMoniker *moniker,
			  Bonobo_Moniker parent,
			  const CORBA_char *name,
			  CORBA_Environment *ev)
{
	int i;

	g_return_val_if_fail(moniker != NULL, CORBA_OBJECT_NIL);
	g_return_val_if_fail(strlen(name) >= PREFIX_LEN, CORBA_OBJECT_NIL);

	bonobo_moniker_set_parent(moniker, parent, ev);

	i = bonobo_moniker_util_seek_std_separator(name, 0);

	bonobo_moniker_set_name(moniker, name, i);

	return bonobo_moniker_util_new_from_name_full(
		bonobo_object_corba_objref(BONOBO_OBJECT(moniker)),
		&name[i], ev);
}

/* Count number of bytes to skip at start of buf */
static int gz_magic [2] = {0x1f, 0x8b};
/* gzip flag byte */
#define GZ_ASCII_FLAG   0x01 /* bit 0 set: file probably ascii text */
#define GZ_HEAD_CRC     0x02 /* bit 1 set: header CRC present */
#define GZ_EXTRA_FIELD  0x04 /* bit 2 set: extra field present */
#define GZ_ORIG_NAME    0x08 /* bit 3 set: original file name present */
#define GZ_COMMENT      0x10 /* bit 4 set: file comment present */
#define GZ_RESERVED     0xE0 /* bits 5..7: reserved */

static int
count_gzip_header (guint8 *buf, guint32 input_length)
{
	int method, flags;
	guint8 *s = buf;
	guint32 left_len = input_length;
	
	if (left_len < 4)
		return -1;

	if (*s++ != gz_magic [0] || *s++ != gz_magic[1])
		return -2;
	
	method = *s++;
	flags = *s++;
	left_len -= 4;
	
	if (method != Z_DEFLATED || (flags & GZ_RESERVED) != 0) {
		/* If it's not deflated, or the reserved isn't 0 */
		return -3;
	}
	
	/* Skip time, xflags, OS code */
	if (left_len < 6)
		return -4;
	s += 6;
	left_len -= 6;
	
	if (flags & GZ_EXTRA_FIELD) {
		unsigned int len;
		if (left_len < 2)
			return -5;
		len = (unsigned int) (*s++);
		len += ((unsigned int) (*s++)) << 8;
		if (left_len < len)
			return -6;
		s += len;
		left_len -= len;
	}
	
	/* Skip filename */
	if (flags & GZ_ORIG_NAME) {
		while (--left_len != 0 && *s++ != '\0') ;
		if (left_len == 0)
			return -7;
	}
	/* Skip comment */
	if (flags & GZ_COMMENT) {
		while (--left_len != 0 && *s++ != '\0') ;
		if (left_len == 0)
			return -7;
	}
	/* Skip CRC */
	if (flags & GZ_HEAD_CRC) {
		if (left_len < 2)
			return -7;
		s += 2;
		left_len -= 2;
	}
	
	return input_length - left_len;
}

static char *
uncompress_memory (char *input_buffer, int size, int *out_size)
{
	z_stream zs;
	gchar *outbuf = NULL;
	GByteArray *data;
	int data_len = 0;
	int zret;
	int gzip_hdr;

	g_return_val_if_fail(input_buffer, NULL);

	data = g_byte_array_new ();

	gzip_hdr = count_gzip_header (input_buffer, size);
	if (gzip_hdr < 0) {
		/* This is not a gzipped stream. Just return the stuff */
		g_byte_array_free(data, FALSE);
		outbuf = g_memdup(input_buffer, size);
		if (out_size)
			*out_size = size;
		return outbuf;
	}

	zs.next_in = input_buffer + gzip_hdr;
	zs.avail_in = size - gzip_hdr;
	zs.zalloc = NULL;
	zs.zfree = NULL;
	zs.opaque = NULL;
	
	outbuf = g_malloc (10000);
	zs.next_out = outbuf;
	zs.avail_out = 10000;

	/* Negative inflateinit is magic to tell zlib that there is no
	 * zlib header */
	inflateInit2 (&zs, -MAX_WBITS);

	while (1) {
		zret = inflate (&zs, Z_SYNC_FLUSH);
		if (zret != Z_OK && zret != Z_STREAM_END)
			break;

		g_byte_array_append(data, outbuf, 10000 - zs.avail_out);
		data_len += 10000 - zs.avail_out;
		zs.next_out = outbuf;
		zs.avail_out = 10000;

		if (zret == Z_STREAM_END)
			break;
	}
	
	if (zret != Z_STREAM_END)
		g_warning ("libz inflate failed! (%d)", zret);
	
	inflateEnd (&zs);

	g_free (outbuf);

	outbuf = data->data;
	if (out_size)
		*out_size = data->len;
	g_byte_array_free(data, FALSE);

	return outbuf;
}


static Bonobo_Unknown
gunzip_resolve (BonoboMoniker *moniker,
		const Bonobo_ResolveOptions *options,
		const CORBA_char *requested_interface,
		CORBA_Environment *ev)
{
	Bonobo_Moniker parent;
	const char *parent_name;
	Bonobo_Stream in_stream;
	BonoboStream *memstream;
	Bonobo_Stream out_stream;
	int len;
	Bonobo_Stream_iobuf *buffer;
	gboolean cont;
	char *ob;
	GByteArray *data;

	g_warning ("Going to resolve the gunzip now");

	parent = bonobo_moniker_get_parent (moniker, ev);
	parent_name = bonobo_moniker_client_get_name (parent, ev);

	if (ev->_major != CORBA_NO_EXCEPTION)
		return CORBA_OBJECT_NIL;

	if (parent == CORBA_OBJECT_NIL) {
		g_warning ("gunzip moniker with no parent!");
		CORBA_exception_set (
			ev, CORBA_USER_EXCEPTION,
			ex_Bonobo_Moniker_InterfaceNotFound, NULL);
		return CORBA_OBJECT_NIL;
	}

	in_stream = Bonobo_Moniker_resolve (
		parent, options, "IDL:Bonobo/Stream:1.0", ev);

	if (ev->_major != CORBA_NO_EXCEPTION)
		goto return_unref_parent;

	if (in_stream == CORBA_OBJECT_NIL) {
		g_warning("Failed to obtain a stream from the parent");
		CORBA_exception_set (
			ev, CORBA_USER_EXCEPTION,
			ex_Bonobo_Moniker_InterfaceNotFound, NULL);
		goto return_unref_parent;
	}

	/* This is necessary because while there is a 
	   bonobo_stream_client_read_string, there isn't any
	   bonobo_stream_client_read, and this compressed data isn't a
	   string. */
	data = g_byte_array_new ();
	do {
		Bonobo_Stream_read (in_stream, 1, &buffer, ev);
		
		if (ev->_major != CORBA_NO_EXCEPTION) {
			g_byte_array_free (data, TRUE);
			goto return_unref_parent;
		}
		cont = buffer->_length ? TRUE : FALSE;
		if (buffer->_length) {
			data = g_byte_array_append (
				data, buffer->_buffer, buffer->_length);
		}
		CORBA_free (buffer);
	} while (cont);

	/* This function will return a memdup of the data we pass to it if it
	   can't decode the header. */
	ob = uncompress_memory (data->data, data->len, &len);
	g_byte_array_free (data, TRUE);

	memstream = bonobo_stream_mem_create (ob, len, TRUE, FALSE);
	g_free (ob);
	out_stream = bonobo_object_corba_objref (BONOBO_OBJECT (memstream));

	if (ev->_major != CORBA_NO_EXCEPTION)
		return CORBA_OBJECT_NIL;

	if (out_stream == CORBA_OBJECT_NIL)
		goto return_unref_parent;

	bonobo_object_release_unref (parent, ev);

	return out_stream;

 return_unref_parent:
	bonobo_object_release_unref (parent, ev);
	return CORBA_OBJECT_NIL;
}

static void
bonobo_moniker_gunzip_class_init(BonoboMonikerGunzipClass *klass)
{
	BonoboMonikerClass *mclass = (BonoboMonikerClass *) klass;

	bonobo_moniker_gunzip_parent_class = gtk_type_class (
		bonobo_moniker_get_type ());

	mclass->parse_display_name = gunzip_parse_display_name;
	mclass->resolve = gunzip_resolve;
}

GtkType
bonobo_moniker_gunzip_get_type(void)
{
	static GtkType type = 0;

	if (!type) {
		GtkTypeInfo info = {
			"BonoboMonikerGunzip",
			sizeof (BonoboMonikerGunzip),
			sizeof (BonoboMonikerGunzipClass),
			(GtkClassInitFunc) bonobo_moniker_gunzip_class_init,
			(GtkObjectInitFunc) NULL,
			NULL, /* reserved 1 */
			NULL, /* reserved 2 */
			(GtkClassInitFunc) NULL
		};

		type = gtk_type_unique (bonobo_moniker_get_type (), &info);
	}

	return type;
}

static BonoboMonikerGunzip *
bonobo_moniker_gunzip_new (void)
{
	BonoboMoniker *moniker;
	Bonobo_Moniker corba_moniker;

	moniker = gtk_type_new (bonobo_moniker_gunzip_get_type ());
	corba_moniker = bonobo_moniker_corba_object_create (
		BONOBO_OBJECT (moniker));

	if (corba_moniker == CORBA_OBJECT_NIL) {
		bonobo_object_unref (BONOBO_OBJECT (moniker));
		return NULL;
	}

	if (!bonobo_object_construct (BONOBO_OBJECT (moniker), corba_moniker)) {
		bonobo_object_unref (BONOBO_OBJECT (moniker));
		return NULL;
	}

	return BONOBO_MONIKER_GUNZIP (moniker);
}

static void
gunzip_destroy_cb (BonoboMoniker *moniker)
{
	running_objects--;

	if (running_objects > 0)
		return;

	bonobo_object_unref (BONOBO_OBJECT (gunzip_factory));
	gtk_main_quit ();
}

static BonoboObject *
bonobo_moniker_gunzip_factory (BonoboGenericFactory *this, void *closure)
{
	BonoboMonikerGunzip *gunzip_moniker = bonobo_moniker_gunzip_new ();

	gtk_signal_connect (
		GTK_OBJECT (gunzip_moniker), "destroy",
		GTK_SIGNAL_FUNC (gunzip_destroy_cb), NULL);

	running_objects++;

	return BONOBO_OBJECT (gunzip_moniker);
}

int
main (int argc, char *argv [])
{
	CORBA_Environment ev;
	CORBA_ORB orb = CORBA_OBJECT_NIL;

	CORBA_exception_init (&ev);

	gnome_init_with_popt_table (
		"gunzip-moniker", "0.0", argc, argv, oaf_popt_options, 0, NULL);
	orb = oaf_init (argc, argv);

	if (!bonobo_init (orb, CORBA_OBJECT_NIL, CORBA_OBJECT_NIL))
		g_error (_("Could not initialize Bonobo"));

	gunzip_factory = bonobo_generic_factory_new (
		"OAFIID:Bonobo_Moniker_gzipFactory",
		bonobo_moniker_gunzip_factory, NULL);

	bonobo_main ();

	CORBA_exception_free (&ev);

	return 0;
}
